# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import datetime
import math
import sys
import time
import traceback

from moirai.database import DatabaseV1
from moirai.hardware.configured_hardware import ConfiguredHardware
from moirai.hardware.timer import Timer


class Free(object):
    __instance = None

    @classmethod
    def instance(cls):
        if not Free.__instance:
            Free.__instance = Free()
        return Free.__instance

    def __init__(self):
        self.last_run = time.time()
        self.running = False
        self.timer = Timer(math.inf, 1)
        self.hardware = None
        self.db = DatabaseV1()
        self.start_time = datetime.datetime.utcnow()
        self.locks = []
        self.inputs = []
        self.outputs = []
        self.graph_id = None

    def is_valid(self):
        return (time.time() - self.last_run) < max(2 * self.timer.interval, 1)

    def run(self, data=None):
        try:
            if data:
                self.last_run = time.time()
                self.running = True
                self.timer.interval = float(data["dt"])
                self.inputs = data["inputs"]
                self.outputs = [
                    out for out in data["outputs"] if len(out["alias"]) != 0
                ]
                config = self.db.get_setting("hardware_configuration")
                self.locks = list(map(self.interlock, config["interlocks"]))

            if not self.is_valid():
                self.running = False

            if self.running:
                if not self.hardware:
                    self.db.set_setting("test_error", "")
                    self.timer = Timer(math.inf, float(data["dt"]))
                    self.start_time = datetime.datetime.utcnow()
                    self.hardware = ConfiguredHardware()
                    self.last_run = time.time()
                    self.graph_id = self.db.save_test("Free", self.start_time)

                    for output in self.outputs:
                        self.db.save_test_sensor_value(
                            self.graph_id, output["alias"], 0, 0
                        )
                    for input in self.inputs:
                        self.db.save_test_sensor_value(self.graph_id, input, 0, 0)

                self.timer.sleep()

                for output in self.outputs:
                    self.hardware.write(output["alias"], output["value"])
                    self.db.save_test_sensor_value(
                        self.graph_id,
                        output["alias"],
                        output["value"],
                        self.timer.elapsed(),
                    )

                for input in self.inputs:
                    value = self.hardware.read(input)
                    self.db.save_test_sensor_value(
                        self.graph_id, input, value, self.timer.elapsed()
                    )

                for lock in self.locks:
                    lock()
            elif self.hardware:
                self.shutdown()
        except Exception as e:
            print(e)
            self.running = False
            self.shutdown()

    def shutdown(self):
        if self.hardware:
            self.off_values = {
                p["alias"]: float(p["defaultValue"])
                for p in self.hardware.ports
                if p["type"] & (8 | 16)
            }
            for k, v in self.off_values.items():
                self.hardware.write(k, v)
            self.hardware = None

    def interlock(self, lock):
        try:
            code = compile("y=%s" % lock["expression"], "_string_", "exec")
        except SyntaxError as err:
            error = err.__class__.__name__
            detail = err.args[0]
            line = err.lineno
            error_string = "%s on %s:%s: %s" % (error, "pid", line, detail)
            print(error_string)
            self.db.set_setting("test_error", error_string)
            self.running = False
            return

        def f(code=code, lock=lock):
            try:
                value = self.hardware.read(lock["sensor"])
                scope = {"x": value}
                exec(code, None, scope)
            except Exception as err:
                error = err.__class__.__name__
                detail = err.args[0]
                cl, exc, tb = sys.exc_info()
                tb = self.stringify_tb(traceback.extract_tb(tb))
                error_string = "%s: %s\n%s" % (error, detail, tb)
                print(error_string)
                self.db.set_setting("test_error", error_string)
                raise Exception("Interlock")
            if scope["y"]:
                self.hardware.write(lock["actuator"], lock["actuatorValue"])
                self.db.set_setting("test_error", "Interlock")
                raise Exception("Interlock")

        return f

    def stringify_tb(self, tb):
        msg = "Traceback:\n\t"
        msg += "\n\t".join(["%s:%s in %s" % (t.filename, t.lineno, t.name) for t in tb])
        return msg



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import datetime

from moirai.database import DatabaseV1
from moirai.hardware.configured_hardware import ConfiguredHardware
from moirai.hardware.timer import Timer


class SystemResponseTest(object):
    def __init__(self, test_id):
        self.db = DatabaseV1()
        tests = self.db.get_setting("system_response_tests")
        self.test = next((t for t in tests if t["id"] == test_id), None)
        if self.test is None:
            raise Exception("Test not found")
        self.hardware = ConfiguredHardware()
        configuration = self.db.get_setting("hardware_configuration")
        self.locks = list(map(self.interlock, configuration["interlock"]))
        self.off_values = {
            p["alias"]: float(p["defaultValue"])
            for p in configuration["ports"]
            if p["type"] & (8 | 16)
        }

    def interlock(self, lock):
        code = compile("y=%s" % lock["expression"], "_string_", "exec")

        def f(code=code, lock=lock):
            value = self.hardware.read(lock["sensor"])
            scope = {"x": value}
            exec(code, None, scope)
            if scope["y"]:
                value = float(lock["actuatorValue"])
                self.hardware.write(lock["actuator"], value)
                self.db.set_setting("test_error", "Interlock")
                raise Exception("Interlock")

        return f

    def run(self):
        self.db.set_setting("current_test", self.test["name"])
        self.db.set_setting("test_error", None)

        for k, v in self.off_values.items():
            self.hardware.write(k, v)

        for o in self.test["fixedOutputs"]:
            self.hardware.write(o["alias"], o["value"])

        run_time = self.test["points"][-1]["x"]
        interval = self.test["logRate"]
        t = Timer(run_time, interval)
        ports = self.test["output"]  # Can be a list.
        ports = [ports] if isinstance(ports, str) else ports
        start_time = datetime.datetime.utcnow()
        last_port_value = 0
        t_elapsed = 0
        graph_id = self.db.save_test(self.test["name"], start_time)

        try:
            while self.db.get_setting("current_test") is not None:
                for lock in self.locks:
                    lock()

                for sensor in self.test["inputs"]:
                    value = self.hardware.read(sensor)
                    self.db.save_test_sensor_value(graph_id, sensor, value, t_elapsed)

                for point in self.test["points"]:
                    if t.elapsed() < point["x"]:
                        for port in ports:
                            self.hardware.write(port, point["y"])
                            self.db.save_test_sensor_value(
                                graph_id, port, point["y"], t_elapsed
                            )
                        last_port_value = point["y"]
                        break

                t.sleep()
                t_elapsed = t.elapsed()

        except Exception as e:
            print(e)
            self.db.set_setting("test_error", str(e))

        for port in ports:
            self.db.save_test_sensor_value(graph_id, port, last_port_value, t.elapsed())

        for o in self.test["afterOutputs"]:
            self.hardware.write(o["alias"], o["value"])

        for k, v in self.off_values.items():
            self.hardware.write(k, v)

        self.db.set_setting("current_test", None)
        self.db.close()



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import datetime

import numpy as np
import math
import threading
import time

import sys
import traceback

from moirai.database import DatabaseV1
from moirai.hardware.configured_hardware import ConfiguredHardware
from moirai.hardware.timer import Timer, Finished


class Controller(object):
    def __init__(self, controller_id):
        self.db = DatabaseV1()
        cs = self.db.get_setting("controllers")
        self.cs = next((c for c in cs if c["id"] == controller_id), None)
        if self.cs is None:
            raise Exception("Controller not found")
        self.hardware = ConfiguredHardware()
        configuration = self.db.get_setting("hardware_configuration")
        self.locks = [self.interlock(lo) for lo in configuration["interlocks"]]
        self.off_values = {
            p["alias"]: float(p["defaultValue"])
            for p in configuration["ports"]
            if p["type"] & (8 | 16)
        }
        self.running = True
        self.lock = threading.Lock()
        self.after = None

    def interlock(self, lock):
        try:
            code = compile("y=%s" % lock["expression"], "_string_", "exec")
        except SyntaxError as err:
            error = err.__class__.__name__
            detail = err.args[0]
            line = err.lineno
            scope = "interlock"
            error_string = "%s on %s:%s: %s" % (error, scope, line, detail)
            print(error_string)
            self.db.set_setting("test_error", error_string)
            self.running = False
            return

        def f(code=code, lock=lock):
            try:
                value = self.hardware.read(lock["sensor"])
                scope = {"x": value}
                exec(code, None, scope)
            except Exception as err:
                error = err.__class__.__name__
                detail = err.args[0]
                cl, exc, tb = sys.exc_info()
                tb = self.stringify_tb(traceback.extract_tb(tb))
                error_string = "%s: %s\n%s" % (error, detail, tb)
                print(error_string)
                self.db.set_setting("test_error", error_string)
                raise Exception("Interlock")
            if scope["y"]:
                self.hardware.write(lock["actuator"], lock["actuatorValue"])
                self.db.set_setting("test_error", "Interlock")
                raise Exception("Interlock")

        return f

    def lock_forever(self):
        interval = float(self.cs["tau"])
        while self.running:
            time.sleep(interval)
            self.lock.acquire()
            try:
                for lock in self.locks:
                    lock()
            except:  # noqa: E722 pylint: disable=E722
                try:
                    self.running = False
                    if self.after is not None:
                        ins = self.cs["inputs"]
                        inputs = {s: self.hardware.read(s) for s in ins}
                        plocals = {
                            "inputs": inputs,
                            "outputs": dict(),
                            "np": np,
                            "math": math,
                        }

                        exec(self.after, plocals, plocals)

                        for actuator, value in plocals["outputs"].items():
                            self.hardware.write(actuator, value)
                except Exception as err:
                    error = err.__class__.__name__
                    detail = err.args[0]
                    cl, exc, tb = sys.exc_info()
                    tb = self.stringify_tb(traceback.extract_tb(tb))
                    error_string = "%s: %s\n%s" % (error, detail, tb)
                    print(error_string)
                    self.db.set_setting("test_error", error_string)
            self.lock.release()

    def run(self):
        self.db.set_setting("current_test", self.cs["name"])
        self.db.set_setting("test_error", None)

        after = None
        self.running = True

        try:
            scope = "before"
            before = compile(self.cs["before"], "before", "exec")
            scope = "controller"
            controller = compile(self.cs["controller"], "controller", "exec")
            scope = "after"
            after = compile(self.cs["after"], "after", "exec")
            self.after = after

            thread = threading.Thread(target=self.lock_forever)
            thread.start()
            thread.isDaemon = True

            inputs = {s: self.hardware.read(s) for s in self.cs["inputs"]}
            plocals = {
                "inputs": inputs,
                "outputs": dict(),
                "s": dict(),
                "dt": float(self.cs["tau"]),
                "np": np,
                "math": math,
            }

            for k, v in self.off_values.items():
                self.hardware.write(k, v)

            exec(before, plocals, plocals)

            state = plocals["s"]

            for k, v in plocals["outputs"].items():
                self.hardware.write(k, v)

            run_time = int(self.cs["runTime"])
            interval = float(self.cs["tau"])
            t = Timer(run_time, interval)
            start_time = datetime.datetime.utcnow()
            time = 0
            graph_id = self.db.save_test(self.cs["name"], start_time)

            while self.db.get_setting("current_test") is not None and self.running:
                self.lock.acquire()
                inputs = {s: self.hardware.read(s) for s in self.cs["inputs"]}
                self.lock.release()

                plocals = {
                    "inputs": inputs,
                    "outputs": dict(),
                    "s": state,
                    "log": dict(),
                    "t": time,
                    "dt": interval,
                    "np": np,
                    "math": math,
                }

                exec(controller, plocals, plocals)

                plocals["log"] = {
                    **plocals["log"],
                    **plocals["inputs"],
                    **plocals["outputs"],
                }

                self.lock.acquire()
                for k, v in plocals["outputs"].items():
                    self.hardware.write(k, v)

                for k, v in plocals["log"].items():
                    self.db.save_test_sensor_value(graph_id, k, v, time)
                self.lock.release()

                state = plocals["s"]

                t.sleep()
                time = t.elapsed()
        except Finished:
            pass
        except SyntaxError as err:
            error = err.__class__.__name__
            detail = err.args[0]
            line = err.lineno
            error_string = "%s on %s:%s: %s" % (error, scope, line, detail)
            print(error_string)
            self.db.set_setting("test_error", error_string)
        except Exception as err:
            error = err.__class__.__name__
            detail = err.args[0]
            cl, exc, tb = sys.exc_info()
            tb = self.stringify_tb(traceback.extract_tb(tb))
            error_string = "%s: %s\n%s" % (error, detail, tb)
            print(error_string)
            self.db.set_setting("test_error", error_string)

        try:
            if after is not None:
                inputs = {s: self.hardware.read(s) for s in self.cs["inputs"]}
                plocals = {"inputs": inputs, "outputs": dict(), "np": np, "math": math}

                exec(after, plocals, plocals)

                for actuator, value in plocals["outputs"].items():
                    self.hardware.write(actuator, value)

            for k, v in self.off_values.items():
                self.hardware.write(k, v)
        except Exception as err:
            error = err.__class__.__name__
            detail = err.args[0]
            cl, exc, tb = sys.exc_info()
            tb = self.stringify_tb(traceback.extract_tb(tb))
            error_string = "%s: %s\n%s" % (error, detail, tb)
            print(error_string)
            self.db.set_setting("test_error", error_string)

        self.db.set_setting("current_test", None)
        self.db.close()
        self.running = False

    def stringify_tb(self, tb):
        msg = "Traceback:\n\t"
        msg += "\n\t".join(["%s:%s in %s" % (t.filename, t.lineno, t.name) for t in tb])
        return msg
